#include "distconfig.h"

#include <iostream>

#include "qlistview.h"
#include "qlistbox.h"
#include "qgroupbox.h"
#include "qlineedit.h"
#include "qspinbox.h"
#include "qcombobox.h"
#include "qcheckbox.h"
#include "qpushbutton.h"
#include "qvalidator.h"
#include "qlabel.h"
#include "qfiledialog.h"
#include "qradiobutton.h"

#include "config.h"
#include "configview.h"
#include "configviewui.h"
#include "controller.h"
#include "dcbackend.h"

ProfileItem::ProfileItem(QListView *parent, ConfigView *confview) : 
  QListViewItem(parent, "New profile"),
  profile(newConfig()), conf(getConfig(profile)), configview(confview),
  active(false)
{
  setRenameEnabled(0,true);
  startRename(0);
}

ProfileItem::ProfileItem(QListView *parent, ConfigView *confview, int profile) :
  QListViewItem(parent, getConfig(profile)->name),
  profile(profile), conf(getConfig(profile)), configview(confview),
  active(false)
{
  setRenameEnabled(0,true);
}

ProfileItem::~ProfileItem()
{
  killConfig(profile);
}

bool ProfileItem::activateProfile()
{
  if (active) return false;

  setActiveConfig(profile);
  active = true;
  updateText();
  configview->updateConfigView(*conf);
  return true;
}

void ProfileItem::deactivateProfile()
{
  active = false;
  updateText();
}

void ProfileItem::updateText()
{
  setText(0,conf->name+(active?" (active)":""));
}

void ProfileItem::okRename(int col)
{
  QListViewItem::okRename(col);
  conf->name = text(0);
  updateText();
}

void ProfileItem::cancelRename(int col)
{
  QListViewItem::cancelRename(col);
  updateText();
}

void ProfileItem::startRename(int col)
{
  setText(0,conf->name);
  QListViewItem::startRename(col);
}

/*********************************/
HubItem::HubItem(QListView *parent, QListViewItem *parent2) : 
  QListViewItem(parent, "New hub"),
  type(AllHubs),
  hub(newHub()), name("New hub"), right(0), left(new HubItem(parent2, this)), active(false)
{
  setRenameEnabled(0,true);
  startRename(0);
}

HubItem::HubItem(QListView *parent, QListViewItem *parent2, QString name, QString hubaddr) :
  QListViewItem(parent, name),
  type(AllHubs),
  hub(newHub()), name(name), right(0), left(new HubItem(parent2, this)), active(false)
{
  DcHubConfig *conf = getHub(hub);
  conf->name = name;
  conf->hubaddr = hubaddr;
  setRenameEnabled(0,true);
}

HubItem::HubItem(QListView *parent, QListViewItem *parent2, QString name, QString hubaddr, HubItem *clone) :
  QListViewItem(parent, name),
  type(AllHubs),
  hub(newHub()), name(name), right(0), left(new HubItem(parent2, this)), active(false)
{
  DcHubConfig *conf = getHub(hub);
  *conf = *getHub(clone->getHubId());
  conf->name = name;
  conf->hubaddr = hubaddr;
  setRenameEnabled(0,true);
}

HubItem::HubItem(QListView *parent, QListViewItem *parent2, HubItem *clone) : 
  QListViewItem(parent, "Clone of "+clone->getName()),
  type(AllHubs),
  hub(newHub()), name("Clone of "+clone->getName()), right(0), left(new HubItem(parent2, this)), active(false)
{
  *getHub(hub) = *getHub(clone->getHubId());
  setRenameEnabled(0,true);
  startRename(0);
}

HubItem::HubItem(QListView *parent, QListViewItem *parent2, int hub) :
  QListViewItem(parent, getHub(hub)->name),
  type(AllHubs),
  hub(hub), name(getHub(hub)->name), right(0), left(new HubItem(parent2, this)), active(false)
{
  setRenameEnabled(0,true);
}

HubItem::HubItem(QListView *parent, HubItem *right, bool created) :
  QListViewItem(parent, right->getName()),
  type(ActiveHubs),
  hub(right->getHubId()), name(right->getName()), right(right), left(0)
{
  setRenameEnabled(0,true);
  if (!created) global_config->activehubs.push_back(hub);
  controllerBackend()->addHub(hub, *getHub(hub));
  right->activateHub(this);
}

HubItem::HubItem(QListViewItem *parent, HubItem *right) :
  QListViewItem(parent, right->getName()),
  type(HubConfigs),
  hub(right->getHubId()), name(right->getName()), right(right), left(0)
{
}

HubItem::~HubItem()
{
  switch (type) {

  case AllHubs: 
    delete right;
    delete left; 
    killHub(hub);
    break;
  case ActiveHubs:
    global_config->activehubs.remove(hub);
    controllerBackend()->killHub(hub);
    right->deactivateHub();
    break;
  case HubConfigs:
    break;
  }
}


void HubItem::activateHub(HubItem *other)
{
  right = other;
  active = true;
}

void HubItem::deactivateHub()
{
  right = 0;
  active = false;
}

void HubItem::configureHub(ConfigView *confview)
{
  confview->configureHub(hub, getHub(hub));
}

void HubItem::rename(QString newname)
{
  name = newname;
  setText(0, name);
}

void HubItem::okRename(int col)
{
  QListViewItem::okRename(col);
  name = getHub(hub)->name = text(0);
  if (right) right->rename(name);
  if (left) left->rename(name);
}

void HubItem::cancelRename(int col)
{
  QListViewItem::cancelRename(col);
  setText(0, name);
}

/*********************************/
ConfigView::ConfigView(QWidget *parent)
  : ConfigViewUI(parent), current(profileBox), lastactive(0),
    hubsgetter(new HubListGetter(publichubslist))
{
  confSelector->setSorting(-1);
  new QListViewItem(confSelector, "Trimming");
  new QListViewItem(confSelector, "Connect options");
  new QListViewItem(confSelector, "Viewing config");
  new QListViewItem(confSelector, "General config");
  new QListViewItem(confSelector, "Hubs");
  new QListViewItem(confSelector, "Profiles");
  
  publichubslist->setColumnWidthMode(0,QListView::Manual);
  publichubslist->setColumnWidthMode(1,QListView::Manual);
  publichubslist->setColumnWidthMode(2,QListView::Manual);
  publichubslist->setColumnWidthMode(3,QListView::Manual);

  publichubslist->setColumnWidth(0,100);
  publichubslist->setColumnWidth(1,60);
  publichubslist->setColumnWidth(2,100);
  publichubslist->setColumnWidth(3,100);

  connect(confSelector, SIGNAL(selectionChanged(QListViewItem*)),
	  this, SLOT(selectionChanged(QListViewItem*)));

  connect(viewdetailscheck,SIGNAL(clicked()),this,SLOT(setViewdetails()));

  connect(downloaddiredit,SIGNAL(lostFocus()),this,SLOT(setDownloaddir()));
  connect(downloaddirbtn,SIGNAL(clicked()),this,SLOT(browseDownloaddir()));
  connect(fileformatedit,SIGNAL(lostFocus()),this,SLOT(setFileformat()));
#ifdef HAVE_ID3_TAG_H
  connect(useid3v2check,SIGNAL(toggled(bool)),this,SLOT(setUseId3v2(bool)));
  connect(setid3check,SIGNAL(toggled(bool)),this,SLOT(setSetId3(bool)));
#else
  setid3check->setEnabled(false);
  useid3v2check->setEnabled(false);
#endif
  connect(rejectid3check,SIGNAL(toggled(bool)),this,SLOT(setRejectId3(bool)));
  connect(tempdiredit,SIGNAL(lostFocus()),this,SLOT(setTempdir()));
  connect(tempdirbtn,SIGNAL(clicked()),this,SLOT(browseTempdir()));
  connect(tempprefixedit,SIGNAL(lostFocus()),this,SLOT(setTempprefix()));
  connect(reqartistcheck,SIGNAL(toggled(bool)),this,SLOT(setReqArtist(bool)));
  connect(reqalbumcheck,SIGNAL(toggled(bool)),this,SLOT(setReqAlbum(bool)));
  connect(matchcharsedit,SIGNAL(lostFocus()),this,SLOT(setMatchchars()));
  connect(minbpsedit,SIGNAL(lostFocus()),this,SLOT(setMinbps()));
  connect(maxbpsedit,SIGNAL(lostFocus()),this,SLOT(setMaxbps()));
  
  connect(delayspin,SIGNAL(valueChanged(int)),this,SLOT(setDelay(int)));
  connect(dlcandspin,SIGNAL(valueChanged(int)),this,SLOT(setDlcand(int)));
  connect(canddelayedit,SIGNAL(lostFocus()),this,SLOT(setCanddelay()));
  connect(candtimeoutedit,SIGNAL(lostFocus()),this,SLOT(setCandtimeout()));
  connect(candtryoutedit,SIGNAL(lostFocus()),this,SLOT(setCandtryout()));
  connect(maxqueuespin,SIGNAL(valueChanged(int)),this,SLOT(setMaxqueue(int)));
  connect(maxgdlqueuespin,SIGNAL(valueChanged(int)),this,SLOT(setMaxgdlqueue(int)));
  connect(nosourcetimeoutedit,SIGNAL(lostFocus()),this,SLOT(setNosourcetimeout()));
  connect(stalledtimeoutedit,SIGNAL(lostFocus()),this,SLOT(setStalledtimeout()));
  connect(searchtimeoutedit,SIGNAL(lostFocus()),this,SLOT(setSearchtimeout()));
  connect(gdlnosourcetimeoutedit,SIGNAL(lostFocus()),this,SLOT(setGdlnosourcetmimeout()));
  connect(gdlstalltimeoutedit,SIGNAL(lostFocus()),this,SLOT(setGdlstalltimeout()));
  connect(gdlsourcestalltimeoutedit,SIGNAL(lostFocus()),this,SLOT(setGdlsourcestalltimeout()));
  connect(gdlsourceattemptsspin,SIGNAL(valueChanged(int)),this,SLOT(setGdlsourceattempts(int)));

  connect(hubaddredit,SIGNAL(lostFocus()),this,SLOT(setHubaddr()));
  connect(nickedit,SIGNAL(lostFocus()),this,SLOT(setNick()));
  connect(descedit,SIGNAL(lostFocus()),this,SLOT(setDesc()));
  connect(emailedit,SIGNAL(lostFocus()),this,SLOT(setEmail()));
  connect(conncombo,SIGNAL(highlighted(const QString &)),this,SLOT(setConn(const QString &)));
  connect(usesizeoffscheck,SIGNAL(clicked()),this,SLOT(setUseSizeoffs()));
  connect(sizeoffsedit,SIGNAL(lostFocus()),this,SLOT(setSizeoffs()));
  connect(sizeoffscombo,SIGNAL(highlighted(const QString &)),this,SLOT(setSizeoffsFactor(const QString &)));
  connect(usevsharecheck,SIGNAL(clicked()),this,SLOT(setUseVshare()));
  connect(vsharediredit,SIGNAL(lostFocus()),this,SLOT(setVsharedir()));
  connect(maxuplspin,SIGNAL(valueChanged(int)),this,SLOT(setMaxupl(int)));
  connect(useullimcheck,SIGNAL(clicked()),this,SLOT(setUseUllim()));
  connect(ullimedit,SIGNAL(lostFocus()),this,SLOT(setUllim()));

  QValidator *v = new QIntValidator(0,2000000000,this);

  minbpsedit->setValidator(v);
  maxbpsedit->setValidator(v);
  canddelayedit->setValidator(v);
  candtimeoutedit->setValidator(v);
  candtryoutedit->setValidator(v);
  nosourcetimeoutedit->setValidator(v);
  stalledtimeoutedit->setValidator(v);
  searchtimeoutedit->setValidator(v);
  gdlnosourcetimeoutedit->setValidator(v);
  gdlstalltimeoutedit->setValidator(v);
  gdlsourcestalltimeoutedit->setValidator(v);

  sizeoffsedit->setValidator(v);
  ullimedit->setValidator(v);

  /* Connection options */
  connect(activemoderadio,SIGNAL(clicked()),this,SLOT(setActiveMode()));
  connect(passivemoderadio,SIGNAL(clicked()),this,SLOT(setPassiveMode()));

  connect(ovripcheck,SIGNAL(toggled(bool)),this,SLOT(setUseOverrideIp(bool)));
  connect(ovripedit,SIGNAL(lostFocus()),this,SLOT(setOverideIp()));
  connect(ovrportcheck,SIGNAL(toggled(bool)),this,SLOT(setUseOverridePort(bool)));
  connect(ovrportedit,SIGNAL(lostFocus()),this,SLOT(setOverridePort()));
  connect(dynipcheck,SIGNAL(toggled(bool)),this,SLOT(setDynamicIp(bool)));

  connect(useproxycheck,SIGNAL(toggled(bool)),this,SLOT(setUseProxy(bool)));
  connect(proxyvercombo,SIGNAL(highlighted(const QString &)),this,SLOT(setProxyVersion(const QString &)));
  connect(proxyaddredit,SIGNAL(lostFocus()),this,SLOT(setProxyAddr()));
  connect(proxyportedit,SIGNAL(lostFocus()),this,SLOT(setProxyPort()));
  connect(proxyuseredit,SIGNAL(lostFocus()),this,SLOT(setProxyUser()));
  connect(proxypassedit,SIGNAL(lostFocus()),this,SLOT(setProxyPass()));

  connect(hubsnoproxyradio,SIGNAL(clicked()),this,SLOT(setHubsNoProxy()));
  connect(hubssocksradio,SIGNAL(clicked()),this,SLOT(setHubsSocksProxy()));
  connect(hubswebradio,SIGNAL(clicked()),this,SLOT(setHubsWebProxy()));

  connect(hubsproxyaddredit,SIGNAL(lostFocus()),this,SLOT(setHubsProxyAdress()));
  connect(hubsproxyportedit,SIGNAL(lostFocus()),this,SLOT(setHubsProxyPort()));

  proxyportedit->setValidator(v);
  hubsproxyportedit->setValidator(v);

  /* Controls */
  connect(profileActivateButton,SIGNAL(clicked()),this,SLOT(profileActivate()));
  connect(profileNewButton,SIGNAL(clicked()),this,SLOT(profileNew()));
  connect(profileDeleteButton,SIGNAL(clicked()),this,SLOT(profileDelete()));

  connect(hubsGetButton,SIGNAL(clicked()),hubsgetter,SLOT(getHubs()));
  connect(hubsAddPubButton,SIGNAL(clicked()),this,SLOT(hubsAddPub()));
  connect(hubsAddCloneButton,SIGNAL(clicked()),this,SLOT(hubsAddClone()));
  connect(hubsNewButton,SIGNAL(clicked()),this,SLOT(hubsNew()));
  connect(hubsCloneButton,SIGNAL(clicked()),this,SLOT(hubsClone()));
  connect(hubsDeleteButton,SIGNAL(clicked()),this,SLOT(hubsDelete()));
  connect(hubsActivateButton,SIGNAL(clicked()),this,SLOT(hubsActivate()));
  connect(hubsDeactivateButton,SIGNAL(clicked()),this,SLOT(hubsDeactivate()));
  connect(hubsApplyButton,SIGNAL(clicked()),this,SLOT(hubsApply()));

  connect(sharesAddButton,SIGNAL(clicked()),this,SLOT(sharesAdd()));
  connect(sharesDeleteButton,SIGNAL(clicked()),this,SLOT(sharesDelete()));


  /* Hubs list */
  QValueList<int>::iterator it;
  QValueList<int> hubs = listHubs();
  for (it = hubs.begin(); it != hubs.end(); ++it) {
    new HubItem(hubslist, confSelector->findItem("Hubs",0), *it);
  }

  QValueList<int> profiles = listConfigs();
  int active = getActiveConfig();
  for (it = profiles.begin(); it != profiles.end(); ++it) {
    int id = *it;
    ProfileItem *p = new ProfileItem(profileList,this,id);
    if (id == active) {
      p->activateProfile();
      lastactive = p;
    }
  }
}

ConfigView::~ConfigView()
{
}

void ConfigView::updateConfigView(const Config &newconf)
{
  QString numstr;
  viewdetailscheck->setChecked(newconf.viewdetails);
  downloaddiredit->setText(newconf.downloaddir);
  fileformatedit->setText(newconf.fileformat);
  useid3v2check->setChecked(newconf.useid3v2);
  setid3check->setChecked(newconf.setid3);
  useid3v2check->setEnabled(newconf.setid3);
  rejectid3check->setChecked(newconf.rejectid3);
  tempdiredit->setText(newconf.tempdir);
  tempprefixedit->setText(newconf.tempprefix);
  reqartistcheck->setChecked(newconf.reqartist);
  reqalbumcheck->setChecked(newconf.reqalbum);
  matchcharsedit->setText(newconf.matchchars);
  minbpsedit->setText(numstr.setNum(newconf.minbps));
  maxbpsedit->setText(numstr.setNum(newconf.maxbps));

  delayspin->setValue(newconf.delay);
  dlcandspin->setValue(newconf.dlcand);
  canddelayedit->setText(numstr.setNum(newconf.canddelay));
  candtimeoutedit->setText(numstr.setNum(newconf.candtimeout));
  candtryoutedit->setText(numstr.setNum(newconf.candtryout));
  maxqueuespin->setValue(newconf.maxqueue);
  maxgdlqueuespin->setValue(newconf.maxgdlqueue);
  nosourcetimeoutedit->setText(numstr.setNum(newconf.nosourcetimeout));
  stalledtimeoutedit->setText(numstr.setNum(newconf.stalledtimeout));
  searchtimeoutedit->setText(numstr.setNum(newconf.searchtimeout));
  gdlnosourcetimeoutedit->setText(numstr.setNum(newconf.gdlnosourcetimeout));
  gdlstalltimeoutedit->setText(numstr.setNum(newconf.gdlstalltimeout));
  gdlsourcestalltimeoutedit->setText(numstr.setNum(newconf.gdlsourcestalltimeout));
  gdlsourceattemptsspin->setValue(newconf.gdlsourceattempts);
  
  /* Connection options */
  activemoderadio->setChecked(newconf.activemode);
  passivemoderadio->setChecked(!newconf.activemode);

  ovripcheck->setChecked(newconf.useoverrideip);
  ovripedit->setText(newconf.overrideip);
  if (newconf.useoverrideip) {
    dynipcheck->setEnabled(false);
  }
  else {
    ovripedit->setEnabled(false);
  }
  ovrportcheck->setChecked(newconf.useoverrideport);
  ovrportedit->setText(QString::number(newconf.overrideport));
  if (!newconf.useoverrideport) {
    ovrportedit->setEnabled(false);
  }
  dynipcheck->setChecked(newconf.dynamicip && !newconf.useoverrideip);
  if (newconf.dynamicip && !newconf.useoverrideip) {
    ovripcheck->setEnabled(false);
    ovripedit->setEnabled(false);
  }
  useproxycheck->setChecked(newconf.useproxy);
  proxyvercombo->setEnabled(newconf.useproxy);
  proxyaddredit->setEnabled(newconf.useproxy);
  proxyportedit->setEnabled(newconf.useproxy);
  proxyuseredit->setEnabled(newconf.useproxy);
  proxypassedit->setEnabled(newconf.useproxy);

  if (newconf.proxyversion == SOCKS4)
    proxyvercombo->setCurrentText("SOCKS v4");
  else if (newconf.proxyversion == SOCKS5)
    proxyvercombo->setCurrentText("SOCKS v5");

  proxyaddredit->setText(newconf.proxyaddr);
  proxyportedit->setText(QString::number(newconf.proxyport));
  proxyuseredit->setText(newconf.proxyuser);
  proxypassedit->setText(newconf.proxypass);

  switch (newconf.hubsproxytype) {
  case NoProxy:    hubsnoproxyradio->setChecked(true); break;
  case SOCKSProxy: hubssocksradio->setChecked(true); break;
  case WebProxy:   hubswebradio->setChecked(true); break;
  }
  hubsproxyaddredit->setEnabled(newconf.hubsproxytype == WebProxy);
  hubsproxyportedit->setEnabled(newconf.hubsproxytype == WebProxy);


  hubsproxyaddredit->setText(newconf.hubsproxyaddr);
  hubsproxyportedit->setText(QString::number(newconf.hubsproxyport));

  /* Active hubs */
  activehubslist->clear();
  QValueList<int>::const_iterator it;
  for (it = newconf.activehubs.begin(); it != newconf.activehubs.end(); ++it) {
    HubItem *th = dynamic_cast<HubItem*>(hubslist->firstChild());
    while (th) {
      if (th->getHubId() == *it) {
	new HubItem(activehubslist,th,true);
	break;
      }
      th = dynamic_cast<HubItem*>(th->nextSibling());
    }
    if (!th) std::cerr << "Active hub without a real hub!\n";
  }
}

void ConfigView::configureHub(int hub, DcHubConfig *hubconfig)
{
  selected_hub = hubconfig;
  selected_hub_id = hub;

  hubnamelbl->setText("Hub configuration for "+hubconfig->name);

  hubaddredit->setText(hubconfig->hubaddr);
  nickedit->setText(hubconfig->nick);
  descedit->setText(hubconfig->desc);
  emailedit->setText(hubconfig->email);
  conncombo->setCurrentText(hubconfig->conn);

  shareslist->clear();
  QStringList::iterator it;
  for (it = hubconfig->shareddirs.begin(); it != hubconfig->shareddirs.end(); ++it) {
    new QListViewItem(shareslist, *it);
  }
  usesizeoffscheck->setChecked(hubconfig->usesizeoffs);
  double sizeoffs = hubconfig->sizeoffs;
  QString tmp;
  if (sizeoffs >= 1.0*1024*1024*1024*1024) {
    sizeoffscombo->setCurrentText("TB");
    sizeoffsfactor = 1.0*1024*1024*1024*1024;
    tmp.setNum(static_cast<unsigned long>(sizeoffs / (1.0*1024*1024*1024*1024)));
  }
  else if (sizeoffs >= 1.0*1024*1024*1024) {
    sizeoffscombo->setCurrentText("GB");
    sizeoffsfactor = 1.0*1024*1024*1024;
    tmp.setNum(static_cast<unsigned long>(sizeoffs / (1.0*1024*1024*1024)));
  }
  else if (sizeoffs >= 1.0*1024*1024) {
    sizeoffscombo->setCurrentText("MB");
    sizeoffsfactor = 1.0*1024*1024;
    tmp.setNum(static_cast<unsigned long>(sizeoffs / (1.0*1024*1024)));
  }
  else if (sizeoffs >= 1.0*1024) {
    sizeoffscombo->setCurrentText("KB");
    sizeoffsfactor = 1.0*1024;
    tmp.setNum(static_cast<unsigned long>(sizeoffs / (1.0*1024)));
  }
  else {
    sizeoffscombo->setCurrentText("bytes");
    sizeoffsfactor = 1.0;
    tmp.setNum(static_cast<unsigned long>(sizeoffs));
  }
  sizeoffsedit->setText(tmp);
  sizeoffsedit->setEnabled(hubconfig->usesizeoffs);
  usevsharecheck->setChecked(hubconfig->usevshare);
  vsharediredit->setText(hubconfig->vsharedir);
  vsharediredit->setEnabled(hubconfig->usevshare);
  maxuplspin->setValue(hubconfig->maxupl);
  useullimcheck->setChecked(hubconfig->useullim);
  ullimedit->setText(QString::number(hubconfig->ullim));
  ullimedit->setEnabled(hubconfig->useullim);
}
  

void ConfigView::selectionChanged(QListViewItem *item)
{
  current->hide();
  QString what = item->text(0);
  if (what == "Profiles") current = profileBox;
  else if (what == "Hubs") current = hubSelectionBox;
  else if (what == "General config") current = generalBox;
  else if (what == "Viewing config") current = viewBox;
  else if (what == "Connect options") current = connectBox;
  else if (what == "Trimming") current = trimBox;
  else {
    current = hubConfigBox;
    dynamic_cast<HubItem*>(item)->configureHub(this);
  }
  current->show();
}

void ConfigView::show()
{
  ConfigViewUI::show();
  /* Then hide the panels we don't want right now */
  profileBox->hide();
  generalBox->hide();
  trimBox->hide();
  hubConfigBox->hide();
  hubSelectionBox->hide();
  viewBox->hide();
  connectBox->hide();
  current->show();
}

void ConfigView::setViewdetails()
{
  global_config->viewdetails = viewdetailscheck->isChecked();
}
void ConfigView::setDownloaddir()
{
  global_config->downloaddir = downloaddiredit->text();
}
void ConfigView::browseDownloaddir()
{
  QString newpath = QFileDialog::getExistingDirectory();
  if (newpath) {
    downloaddiredit->setText(newpath);
    setDownloaddir();
  }
}
void ConfigView::setFileformat()
{
  global_config->fileformat = fileformatedit->text();
}
void ConfigView::setUseId3v2(bool on)
{
  global_config->useid3v2 = on;
}
void ConfigView::setSetId3(bool on)
{
  global_config->setid3 = on;
  useid3v2check->setEnabled(on);
}
void ConfigView::setRejectId3(bool on)
{
  global_config->rejectid3 = on;
}
void ConfigView::setTempdir()
{
  global_config->tempdir = tempdiredit->text();
}
void ConfigView::browseTempdir()
{
  QString newpath = QFileDialog::getExistingDirectory();
  if (newpath) {
    tempdiredit->setText(newpath);
    setTempdir();
  }
}
void ConfigView::setTempprefix()
{
  global_config->tempprefix = tempprefixedit->text();
}
void ConfigView::setReqArtist(bool on)
{
  global_config->reqartist = on;
}
void ConfigView::setReqAlbum(bool on)
{
  global_config->reqalbum = on;
}
void ConfigView::setMatchchars()
{
  global_config->matchchars = matchcharsedit->text();
}
void ConfigView::setMinbps()
{
  global_config->minbps = minbpsedit->text().toLong();
}
void ConfigView::setMaxbps()
{
  global_config->maxbps = maxbpsedit->text().toLong();
}
void ConfigView::setDelay(int value)
{
  global_config->delay = value;
}
void ConfigView::setDlcand(int value)
{
  global_config->dlcand = value;
}
void ConfigView::setCanddelay()
{
  global_config->canddelay = canddelayedit->text().toInt();
}
void ConfigView::setCandtimeout()
{
  global_config->candtimeout = candtimeoutedit->text().toInt();
}
void ConfigView::setCandtryout()
{
  global_config->candtryout = candtryoutedit->text().toInt();
}
void ConfigView::setMaxqueue(int value)
{
  global_config->maxqueue = value;
}
void ConfigView::setMaxgdlqueue(int value)
{
  global_config->maxgdlqueue = value;
}
void ConfigView::setNosourcetimeout()
{
  global_config->nosourcetimeout = nosourcetimeoutedit->text().toInt();
}
void ConfigView::setStalledtimeout()
{
  global_config->stalledtimeout = stalledtimeoutedit->text().toInt();
}
void ConfigView::setSearchtimeout()
{
  global_config->searchtimeout = searchtimeoutedit->text().toInt();
}
void ConfigView::setGdlnosourcetmimeout()
{
  global_config->gdlnosourcetimeout = gdlnosourcetimeoutedit->text().toInt();
}
void ConfigView::setGdlstalltimeout()
{
  global_config->gdlstalltimeout = gdlstalltimeoutedit->text().toInt();
}
void ConfigView::setGdlsourcestalltimeout()
{
  global_config->gdlsourcestalltimeout = gdlsourcestalltimeoutedit->text().toInt();
}
void ConfigView::setGdlsourceattempts(int value)
{
  global_config->gdlsourceattempts = value;
}

void ConfigView::setHubaddr()
{
  selected_hub->hubaddr = hubaddredit->text();
}
void ConfigView::setNick()
{
  selected_hub->nick = nickedit->text();
}
void ConfigView::setDesc()
{
  selected_hub->desc = descedit->text();
}
void ConfigView::setEmail()
{
  selected_hub->email = emailedit->text();
}
void ConfigView::setConn(const QString &value)
{
  selected_hub->conn = value;
}
void ConfigView::setUseSizeoffs()
{
  selected_hub->usesizeoffs = usesizeoffscheck->isChecked();
  sizeoffsedit->setEnabled(usesizeoffscheck->isChecked());
}
void ConfigView::setSizeoffs()
{
  QString tmp;
  double sizeoffs = sizeoffsfactor*sizeoffsedit->text().toDouble();

  if (sizeoffs >= 1.0*1024*1024*1024*1024) {
    sizeoffscombo->setCurrentText("TB");
    setSizeoffsFactor("TB");
    tmp.setNum(static_cast<unsigned long>(sizeoffs / (1.0*1024*1024*1024*1024)));
  }
  else if (sizeoffs >= 1.0*1024*1024*1024) {
    sizeoffscombo->setCurrentText("GB");
    setSizeoffsFactor("GB");
    tmp.setNum(static_cast<unsigned long>(sizeoffs / (1.0*1024*1024*1024)));
  }
  else if (sizeoffs >= 1.0*1024*1024) {
    sizeoffscombo->setCurrentText("MB");
    setSizeoffsFactor("MB");
    tmp.setNum(static_cast<unsigned long>(sizeoffs / (1.0*1024*1024)));
  }
  else if (sizeoffs >= 1.0*1024) {
    sizeoffscombo->setCurrentText("KB");
    setSizeoffsFactor("KB");
    tmp.setNum(static_cast<unsigned long>(sizeoffs / (1.0*1024)));
  }
  else {
    sizeoffscombo->setCurrentText("bytes");
    setSizeoffsFactor("bytes");
    tmp.setNum(static_cast<unsigned long>(sizeoffs));
  }
  sizeoffsedit->setText(tmp);
  selected_hub->sizeoffs = sizeoffs;
}
void ConfigView::setSizeoffsFactor(const QString &f)
{
  if (f == "bytes") sizeoffsfactor = 1.0;
  if (f == "KB") sizeoffsfactor = 1.0*1024;
  if (f == "MB") sizeoffsfactor = 1.0*1024*1024;
  if (f == "GB") sizeoffsfactor = 1.0*1024*1024*1024;
  if (f == "TB") sizeoffsfactor = 1.0*1024*1024*1024*1024;
  selected_hub->sizeoffs = sizeoffsfactor*sizeoffsedit->text().toDouble();
}
void ConfigView::setUseVshare()
{
  selected_hub->usevshare = usevsharecheck->isChecked();
  vsharediredit->setEnabled(usevsharecheck->isChecked());
}
void ConfigView::setVsharedir()
{
  selected_hub->vsharedir = vsharediredit->text();
}
void ConfigView::setMaxupl(int value)
{
  selected_hub->maxupl = value;
}
void ConfigView::setUseUllim()
{
  selected_hub->useullim = useullimcheck->isChecked();
  ullimedit->setEnabled(useullimcheck->isChecked());
}
void ConfigView::setUllim()
{
  selected_hub->ullim = ullimedit->text().toLong();
}
/* Connection options */
void ConfigView::setActiveMode() 
{
  global_config->activemode = true;
}
void ConfigView::setPassiveMode() 
{
  global_config->activemode = false;
}
void ConfigView::setUseOverrideIp(bool on) 
{
  global_config->useoverrideip = on;
  dynipcheck->setEnabled(!on);
  ovripedit->setEnabled(on);
}
void ConfigView::setOverideIp()
{
  global_config->overrideip = ovripedit->text();
}
void ConfigView::setUseOverridePort(bool on) 
{
  global_config->useoverrideport = on;
  ovrportedit->setEnabled(on);
}
void ConfigView::setOverridePort()
{
  global_config->overrideport = ovrportedit->text().toInt();
}
void ConfigView::setDynamicIp(bool on)
{
  global_config->dynamicip = on;
  ovripcheck->setEnabled(!on);
  ovripedit->setEnabled(!on && global_config->useoverrideip);
}

void ConfigView::setUseProxy(bool on)
{
  global_config->useproxy = on;
  proxyvercombo->setEnabled(on);
  proxyaddredit->setEnabled(on);
  proxyportedit->setEnabled(on);
  proxyuseredit->setEnabled(on);
  proxypassedit->setEnabled(on);
}
void ConfigView::setProxyVersion(const QString &version) 
{
  if (version == "SOCKS v4") global_config->proxyversion=SOCKS4;
  else if (version == "SOCKS v5") global_config->proxyversion=SOCKS5;
}
void ConfigView::setProxyAddr() 
{
  global_config->proxyaddr = proxyaddredit->text();
}
void ConfigView::setProxyPort() 
{
  global_config->proxyport = proxyportedit->text().toInt();
}
void ConfigView::setProxyUser() 
{
  global_config->proxyuser = proxyuseredit->text();
}
void ConfigView::setProxyPass() 
{
  global_config->proxypass = proxypassedit->text();
}

void ConfigView::setHubsNoProxy() 
{
  global_config->hubsproxytype = NoProxy;
  hubsproxyaddredit->setEnabled(false);
  hubsproxyportedit->setEnabled(false);
}
void ConfigView::setHubsSocksProxy() 
{
  global_config->hubsproxytype = SOCKSProxy;
  hubsproxyaddredit->setEnabled(false);
  hubsproxyportedit->setEnabled(false);
}
void ConfigView::setHubsWebProxy() 
{
  global_config->hubsproxytype = WebProxy;
  hubsproxyaddredit->setEnabled(true);
  hubsproxyportedit->setEnabled(true);
}

void ConfigView::setHubsProxyAdress() 
{
  global_config->hubsproxyaddr = hubsproxyaddredit->text();
}
void ConfigView::setHubsProxyPort() 
{
  global_config->hubsproxyport = hubsproxyportedit->text().toInt();
}

/* Controls */
void ConfigView::profileActivate()
{
  if (profileList->selectedItem()) {
    ProfileItem *p = dynamic_cast<ProfileItem*>(profileList->selectedItem());
    if (p->activateProfile()) {
      if (lastactive) lastactive->deactivateProfile();
      lastactive = p;
    }
  }
}

void ConfigView::profileNew()
{
  new ProfileItem(profileList,this);
}

void ConfigView::profileDelete()
{
  if (profileList->childCount() > 1) {
    if (profileList->selectedItem()) {
      ProfileItem *pi = dynamic_cast<ProfileItem*>(profileList->selectedItem());
      bool wasactive = pi->isActive();
      delete pi;
      if (wasactive) {
	lastactive = dynamic_cast<ProfileItem*>(profileList->firstChild());
	lastactive->activateProfile();
      }
    }
  }
}

void ConfigView::hubsAddPub()
{
  QListViewItem *item = publichubslist->selectedItem();
  if (item && !item->text(3).isEmpty())
    new HubItem(hubslist, confSelector->findItem("Hubs",0), item->text(0), item->text(3));
}

void ConfigView::hubsAddClone()
{
  HubItem *clone = dynamic_cast<HubItem*>(hubslist->selectedItem());
  QListViewItem *pubitem = publichubslist->selectedItem();
  if (clone && pubitem && !pubitem->text(3).isEmpty()) {
    new HubItem(hubslist, confSelector->findItem("Hubs",0), pubitem->text(0), pubitem->text(3), clone);
  }
}

void ConfigView::hubsNew()
{
  new HubItem(hubslist, confSelector->findItem("Hubs",0));
}

void ConfigView::hubsClone()
{
  HubItem *clone = dynamic_cast<HubItem*>(hubslist->selectedItem());
  if (clone) new HubItem(hubslist, confSelector->findItem("Hubs",0), clone);
}

void ConfigView::hubsDelete()
{
  delete hubslist->selectedItem();
}

void ConfigView::hubsActivate()
{
  if (hubslist->selectedItem()) {
    HubItem *hi = dynamic_cast<HubItem*>(hubslist->selectedItem());
    if (!hi->isActive())
      new HubItem(activehubslist, hi);
  }
}

void ConfigView::hubsDeactivate()
{
  delete activehubslist->selectedItem();
}

void ConfigView::hubsApply()
{
  controllerBackend()->reconfigureHub(selected_hub_id, *selected_hub);
}

void ConfigView::sharesAdd()
{
  QString dir = QFileDialog::getExistingDirectory();
  if (dir && !shareslist->findItem(dir,ExactMatch)) {
    new QListViewItem(shareslist,dir);
    selected_hub->shareddirs.push_back(dir);
  }
}

void ConfigView::sharesDelete()
{
  QListViewItem *cur = shareslist->selectedItem();
  if (cur) {
    selected_hub->shareddirs.remove(cur->text(0));
    delete cur;
  }
}
    

