#include "qprocess.h"
#include "qregexp.h"
#include "qdict.h"
#include "qintdict.h"
#include "qvaluevector.h"

#include "dcbackend.h"
#include "controller.h"

using namespace std;

/*****************************/
/* DcTransfer implementation */
/*****************************/

DcTransfer::DcTransfer(TransferAgent *agent, QString key) 
  : Transfer(agent, key),
    parent(0), key(key),
    pos(-1), size(-1), progress(-1), target(-1), speed(-1), eta(-1),
    complete(false),
    compsize(0), compcount(0)
{
}

DcTransfer::DcTransfer(DcTransfer *parent, QString key, bool createagent) 
  : Transfer(parent && createagent ? parent->transferChildCreated() : 0, key),
    parent(parent), key(key),
    pos(-1), size(-1), progress(-1), target(-1), speed(-1), eta(-1),
    complete(false),
    compsize(0), compcount(0)
{
  if (parent) parent->addChildTransfer(key, this);
}

DcTransfer::~DcTransfer() 
{ 
  /* Delete your children (may already have been removed,
   * but that's ok, the list is then empty */
  deleteChildren();
  /* Then say goodbye to your parent */
  if (parent) parent->removeChildTransfer(key);
}

void DcTransfer::searchHit(DcSearchResult &result)
{
  /* Default behaivour is to just pass it along, should be overridden */
  QDictIterator<DcTransfer> it(children);
  DcTransfer *cur;
  while ((cur = it.current())) {
    cur->searchHit(result);
    ++it;
  }
}

void DcTransfer::tick(time_t now)
{
  /* Default behaivour is to just pass it along, should be overridden */
  QDictIterator<DcTransfer> it(children);
  DcTransfer *cur;
  while ((cur = it.current())) {
    cur->tick(now);
    ++it;
  }
}

void DcTransfer::updateStatusFromChildren()
{
  if (children.isEmpty() && !compcount) {
    transferStatusChanged("Empty");
    return;
  }

  QDictIterator<DcTransfer> it(children);
  DcTransfer *current;

  long totpos   = 0;
  long totsize  = 0;
  long totspeed = 0;
  long maxeta   = -1;
  double sumprogress = 0;
  double sumtarget = 0;
  int running = 0;
  int total = 0;
  bool allcomplete = true;

  while ((current = it.current())) {
    if (current->getPos() > 0) running++;

    totpos += current->getPos();
    totsize += current->getSize();

    if (!current->isComplete()) {
      if (current->getSpeed() > 0)
	totspeed += current->getSpeed();
      if (current->getProgress() > 0)
	sumprogress += current->getProgress();
      if (current->getTarget() > 0)
	sumtarget += current->getTarget();
      allcomplete = false;
      if (current->getEta() >= 0 && current->getEta() > maxeta) maxeta = current->getEta();
    }
    else {
      sumprogress += 1;
      sumtarget += 1;
    }
    ++it;
    total++;
  }  

  progress = (sumprogress + compcount) / (total + compcount);
  target   = (sumtarget + compcount)   / (total + compcount);
  
  pos = totpos + compsize;
  size = totsize + compsize;
  speed = allcomplete ? -1 : totspeed;
  eta = maxeta;
  complete = allcomplete;

  if (allcomplete) {
    transferStatusChanged("Done!", pos, size, speed, eta, progress, target);
    transferCompleted();
    if (DcTransfer *parent = getParentTransfer()) parent->childCompleted(this);
    delete this;
  }
  else {
    QString status;
    if (speed > 0)     status = "Downloading";
    else if (size > 0) status = "Stalled";
    else               status = "Searching";
    transferStatusChanged(status, pos, size, speed, eta, progress, target);
  }
}

void DcTransfer::childCompleted(DcTransfer *child)
{
  compsize += child->getSize();
  compcount++;
}

bool DcTransfer::startTransfer()
{
  /* Default behaivour is to just pass it along, should be overridden */
  QDictIterator<DcTransfer> it(children);
  DcTransfer *cur;
  while ((cur = it.current())) {
    cur->startTransfer();
    ++it;
  }
  return true;
}

bool DcTransfer::stopTransfer()
{
  /* Default behaivour is to just pass it along, should be overridden */
  QDictIterator<DcTransfer> it(children);
  DcTransfer *cur;
  while ((cur = it.current())) {
    cur->stopTransfer();
    ++it;
  }
  return true;
}

bool DcTransfer::resetTransfer()
{
  /* Default behaivour is to just pass it along, should be overridden */
  QDictIterator<DcTransfer> it(children);
  DcTransfer *cur;
  while ((cur = it.current())) {
    cur->resetTransfer();
    ++it;
  }
  return true;
}


QString DcTransfer::defunchyfy(QString str)
{
  QString modmatchchars = global_config->matchchars;
  modmatchchars.replace("[","\\[");
  modmatchchars.replace("]","\\]");
  modmatchchars.replace("(","\\(");
  modmatchchars.replace(")","\\)");
  str.replace(QRegExp("([^"+global_config->matchchars+"]+)"), " ");
  str.replace(QRegExp("\\s+")," ");
  str.stripWhiteSpace();
  return str;
}

void DcTransfer::deleteChildren()
{  
  QDictIterator<DcTransfer> it(children);
  DcTransfer *current;
  
  while ((current = it.current())) {
    // Deleting this object will remove it from our iterator,
    // so we should *not* increment it!
    delete current; 
  }  
}

/***********************************/
/* DcArtistTransfer implementation */
/***********************************/

DcArtistTransfer::DcArtistTransfer(DcBackend *parent, QString artist) 
  : DcTransfer(parent, artist) {}

DcArtistTransfer::~DcArtistTransfer() { deleteChildren(); }

void DcArtistTransfer::searchHit(DcSearchResult &result)
{
  QString modartist = defunchyfy(key);
  if (!global_config->reqartist || result.match.find(modartist, 0, false) != -1)
    DcTransfer::searchHit(result);
}

void DcArtistTransfer::tick(time_t now)
{
  DcTransfer::tick(now);
  updateStatusFromChildren();
}

/**********************************/
/* DcAlbumTransfer implementation */
/**********************************/

DcAlbumTransfer::DcAlbumTransfer(DcArtistTransfer *parent, QString album) : 
  DcTransfer(parent, album) {}
  

DcAlbumTransfer::~DcAlbumTransfer() { deleteChildren(); }

void DcAlbumTransfer::searchHit(DcSearchResult &result)
{
  QString modalbum = defunchyfy(key);
  if (!global_config->reqalbum || (result.match.find(modalbum, 0, false) != -1))
    DcTransfer::searchHit(result);
}

void DcAlbumTransfer::tick(time_t now)
{
  DcTransfer::tick(now);
  updateStatusFromChildren();
}

/**************************************/
/* DcGDLSourceTransfer implementation */
/**************************************/

DcGDLSourceTransfer::DcGDLSourceTransfer(DcGDLTransfer *parent,
					 QString user, QString path) :
  DcTransfer(parent, user+"|"+path, global_config->viewdetails),
  user(user), path(path), 
  started(false),
  lastpos(0), stalled(false), attempts(0)
{}

DcGDLSourceTransfer::~DcGDLSourceTransfer()
{
  if (started) getBackend()->killGDLSource(getHub(), getGid(), user, path);
  deleteChildren();
}

bool DcGDLSourceTransfer::startTransfer()
{
  if (!started) {
    started = true;
    getBackend()->createGDLSource(getHub(), getGid(), user, path, getSize());
  }
  return true;
}

bool DcGDLSourceTransfer::stopTransfer()
{
  if (started) {
    started = false;
    getBackend()->killGDLSource(getHub(), getGid(), user, path);
  }
  transferStatusChanged("Stopped",pos,size,speed,eta,progress,target);
  return true;
}

bool DcGDLSourceTransfer::resetTransfer()
{
  int hub = getHub();
  int gid = getGid();
  getBackend()->killGDLSource(hub, gid, user, path);
  getBackend()->createGDLSource(hub, gid, user, path, getSize());
  stalled = false;
  return true;
}

void DcGDLSourceTransfer::setStatus(actiontype action, 
				    long pos, long start, long end)
{
  /* GDL source upkeep algorithm:
   * 1. If a download hasn't started for GDLSOURCESTALLTIMEOUT seconds,
   *    it is restarted.
   * 2. If a download still hasn't been moving after GDLSOURCEATTEMPTS
   *    attempts, it is concidered broken and deletes itself.
   */

  if (!started) return;

  QString actionstr;
  switch (action) {
  case Waiting: actionstr = "Waiting"; break;
  case Trying:  actionstr = "Trying";  break;
  case Running: actionstr = "Running"; break;
  }
  this->pos = pos;

  if (action == Waiting || action == Trying || 
      (action == Running && lastpos == pos)) {
    if (!stalled) {
      stalled = true;
      stalledtime = time(0);
    }
    if (time(0) - stalledtime >= global_config->gdlsourcestalltimeout) {
      transferStatusChanged("Timout, restarting!");
      resetTransfer();
      attempts++;
      if (attempts == global_config->gdlsourceattempts) {
	delete this;
	return;
      }
    }
  }
  else {
    stalled = false;
    attempts = 0;
  }

  size = end-start;
  speed = (pos - lastpos) * 1000 / global_config->delay;
  if (size > 0 && pos == size) eta = -1;
  else if (size <= 0) eta = -1;
  else if (speed <= 0) eta = -2;
  else eta = (size - pos) / speed;
  
  if (pos >= 0 && size > 0) progress = static_cast<double>(pos) / size;
  else progress = -1;
  
  target = 1;
  transferStatusChanged(actionstr, pos, size, speed, eta, progress, target);

  lastpos = pos;
}

/********************************/
/* DcGDLTransfer implementation */
/********************************/

DcGDLTransfer::DcGDLTransfer(DcTrackTransfer *parent, int hub, long size) :
  DcTransfer(parent, intToString(hub)+":"+intToString(size), global_config->viewdetails),
  hub(hub), 
  started(false),
  activated(false), hadgdl(false),
  nosources(false), stalled(false),
  lastpos(0), timerstart(time(0)), starttime(time(0)), creationtime(time(0))
{
  this->size = size; /* This belongs to papa, can't init it in init list */
  gid = getBackend()->createGDL(hub, size, this);
}

DcGDLTransfer::~DcGDLTransfer()
{
  deleteChildren();
  getBackend()->killGDL(hub, gid);
  if (started) dynamic_cast<DcTrackTransfer*>(getParentTransfer())->gdlStopped();
  dynamic_cast<DcTrackTransfer*>(getParentTransfer())->unregisterGDL(this);
}

inline QString DcGDLTransfer::intToString(long i)
{
  QString str;
  str.setNum(i);
  return str;
}

void DcGDLTransfer::searchHit(DcSearchResult &result)
{
  if (childrenCount() > global_config->maxgdlqueue) return;
  if ((getChildTransfer(result.user+"|"+result.path))) return;
  DcGDLSourceTransfer *newtrans = 
    new DcGDLSourceTransfer(this, result.user, result.path);
  if (started) newtrans->startTransfer();
}

void DcGDLTransfer::gdlListingStart()
{
  hadgdl = false;
}

void DcGDLTransfer::gdlListingEnd()
{
  if (!hadgdl && activated) {
    transferCompleted();
    gdlDone();
    return;
  }

  if (started && !activated) transferStatusChanged("Not activated");
  if (!activated && (time(0) - creationtime > GDLACTIVATETIMEOUT)) {
    delete this;
  }
}

bool DcGDLTransfer::hasStarted()
{
  return pos > 0;
}

void DcGDLTransfer::setStatus(long pos, long dctcspeed)
{
  /* GDL upkeep algorithm:
   * 1. If this GDL havn't had any sources in GDLNOSOURCETIMEOUT seconds,
   *    it is considered dead and deletes itself.
   * 2. If this GDL has been stalled for GDLSTALLEDTIMEOUT seconds,
   *    it is considered dead and deletes itself.
   */

  static_cast<void>(dctcspeed); /* Warning eater */
  activated = true;
  hadgdl    = true;

  if (!started) return;

  time_t now = time(0);

  if (childrenCount() == 0) {
    if (!nosources) {
      nosources = true;
      nosourcetime = now;
    }
    if (now - nosourcetime >= global_config->gdlnosourcetimeout) {
      //      log("GDL "+QString::number(gid));
      delete this;
      return;
    }
  }
  else nosources = false;

  if (lastpos == pos) {
    if (!stalled) {
      stalled = true;
      stalledtime = now;
    }
    if (now - stalledtime >= global_config->gdlstalltimeout) {
      delete this;
      return;
    }
  }
  else stalled = false;

  /* Start counting for tryouttime instaed of starttime */
  if (pos > 0 && lastpos == 0) {
    resetTimer();
  }
  lastpos = pos;

  if (pos > 0) {
    this->pos = pos;
    if (now - starttime)
      speed = pos / (now - starttime);
    else
      speed = -1;

    if (size > 0 && pos == size) eta = -1;
    else if (size <= 0) eta = -1;
    else if (speed <= 0) eta = -2;
    else eta = (size - pos) / speed;
    if (pos >= 0 && size > 0) progress = static_cast<double>(pos) / size;
    else progress = -1;
    target = 1;
    QString countstr;
    countstr.setNum(childrenCount());
    transferStatusChanged(QString("Downloading... (")+
			  countstr+" source)", pos, size, speed, eta, progress, target);
  }
  else {
    starttime = now;
    QString countstr;
    countstr.setNum(childrenCount());
    transferStatusChanged(QString("Not yet started... (")+
			  countstr+" source)");
  }
}

void DcGDLTransfer::setSourceStatus(QString id, actiontype action,
				    long pos, long start, long end)
{
  DcGDLSourceTransfer *gdls = dynamic_cast<DcGDLSourceTransfer*>(getChildTransfer(id));
  if (gdls) gdls->setStatus(action, pos, start, end);
}

void DcGDLTransfer::resetTimer()
{
  timerstart = time(0);
}

time_t DcGDLTransfer::getTimer()
{
  if (!started) return 0; /* Stopped transfers doesn't get killed */
  return time(0) - timerstart;
}

bool DcGDLTransfer::startTransfer()
{
  if (!started) {
    started = true;
    dynamic_cast<DcTrackTransfer*>(parent)->gdlStarted();
    nosources = false;
    stalled = false;
    lastpos = 0;
    time_t now = time(0);
    timerstart = starttime = now;
  }
  return DcTransfer::startTransfer();
}

bool DcGDLTransfer::stopTransfer()
{
  if (started) {
    started = false;
    dynamic_cast<DcTrackTransfer*>(parent)->gdlStopped();
  }
  transferStatusChanged("Stopped",pos,size,speed,eta,progress,target);
  return DcTransfer::stopTransfer();
}


/**********************************/
/* DcTrackTransfer implementation */
/**********************************/

DcTrackTransfer::DcTrackTransfer(DcAlbumTransfer *parent,
				 QString title, int trackno, int length) :
  DcTransfer(parent, formatKey(title, trackno, length)),
  title(title),
  trackno(trackno), length(length), running(true),
  rungdl(0), best(0), lastsearch(0), newcandidatetime(0)
{
  performSearch();
}

DcTrackTransfer::~DcTrackTransfer()
{
  deleteChildren();
  controllerEndAudioTransfer(getArtist(), getAlbum(), title, trackno, length);
}

inline QString DcTrackTransfer::formatKey(QString title, 
					  int trackno, int length)
{
  QString trackstr;
  QString lenmin;
  QString lensec;
  trackstr.setNum(trackno);
  if (trackno < 10) trackstr = "0"+trackstr;
  lenmin.setNum(length / 60);
  lensec.setNum(length % 60);
  return trackstr+" "+title+" ["+lenmin+":"+(length % 60 < 10 ? "0"+lensec : lensec)+"]";
}

void DcTrackTransfer::searchHit(DcSearchResult &result)
{
  int hub = result.hub;
  long size = result.size;

  // Test that the file has the title, and correct size
  if (result.match.find(defunchyfy(title), 0, false) == -1) return;
  if (size < length*global_config->minbps) return;
  if (result.size > length*global_config->maxbps) return;

  /* Download selection algorithm:
   * 1. Search for sources.
   * 2. Search hits are stored in a search hit queue. If there are more than
   *    MAXQUEUEDHITS search hits, new ones are dropped.
   * 3. A new download from the queue is promoted to candidate download when
   *    there have been less than GDLCANDIDATES running downloads for
   *    NEWCANDIDATEDELAY seconds.
   * 4. Candidate downloads must start in GDLSTARTTIME seconds, and are killed
   *    if not.
   * 5. Candidate downloads then run for GDLTRYOUT seconds before they are 
   *    evaluated. If a downlad is selected as the best, it gets to keep going. 
   *    If not, it is killed.
   * 6. The download with smallest ETA is selected as the best, and get to
   *    keep running.
   * 7. If a new "best" download is selected, the old one gets downgraded 
   *    to candidate and again gets to try out its speed for GDKTRYOUTTIME 
   *    seconds.
   * 8. New searches are performed every SEARCHTIMEOUT seconds if downloads
   *    are running, every STALLEDSEARCHTIMEOUT seconds if there are 
   *    downloads, but they are all stalled, or every NOSOURCESEARHCTIMEOUT
   *    seconds if there are no sources in the search hit queue.
   */

  /* Only add new search hits directly if there is a running GDL
   * that matches it
   */
  QString tmp;
  QString gdlkey;
  tmp.setNum(result.hub);
  gdlkey.setNum(size);
  gdlkey = tmp+":"+gdlkey;
  DcTransfer *gdl;

  /* Check if we have a suitable GDL handler.
   * If not, and if we don't have enough queued downloads already, 
   * create a new (unstarted) GDL, will promote it to
   * a candidate download and start it when appropriate
   */
  if (!(gdl = getChildTransfer(gdlkey))) {
    if (gdlqueue.count() >= static_cast<unsigned>(global_config->maxqueue)) return;

    gdl = new DcGDLTransfer(this, hub, size);
    gdlqueue.push_back(gdlkey);
  }
  /* If we're still here, all tests passed and gdl is valid,
   * pass on our result for possible GDL source creation */
  gdl->searchHit(result);
}

void DcTrackTransfer::gdlDone(DcGDLTransfer *gdl)
{
  /* One download is finished, see if this is accepted.
   * If the download is OK, kill yourself since you are done.
   * If not, kill that transfer and go on as before (should mark 
   * this source so we don't do the same mistake again?
   */
  if (controllerAcceptAudioTransfer(gdl->getFilename(), getArtist(),
				    getAlbum(), title, trackno, length)) {
    /* Put this data in us so updateStatusFromChildren get it */
    pos = gdl->getPos();
    size = gdl->getSize();
    speed = gdl->getSpeed();
    eta = gdl->getEta();
    progress = gdl->getProgress();
    target = gdl->getTarget();
    
    /* Show the user our status */
    getParentTransfer()->childCompleted(this);
    transferStatusChanged("Done!",pos,size,speed,eta,progress,target);
    transferCompleted();
    //    getParentTranfer()->childCompleted(this);
    delete this;
  }
  else delete gdl;
}

void DcTrackTransfer::tick(time_t now)
{
  /* Stopped transfers should not take up any bandwith or CPU time */
  if (!running) {
    transferStatusChanged("Stopped",pos,size,speed,eta,progress,target);
    return;
  }

  bool stalled=true; // Guilty until proven otherwise
  QDictIterator<DcTransfer> it(children);
  DcGDLTransfer *current;

  while (it.current()) {
    current = dynamic_cast<DcGDLTransfer*>(it.current());
    
    // Gather info for later, see if all downloads are stalled
    if (current->getSpeed() > 0) stalled = false;

    // Compete for best download
    if (!best) best = current; // We have no best, set this
    else if (current->getEta() >= 0 && 
	     (current->getEta() < best->getEta() || best->getEta() < 0)) {
      best->resetTimer(); // Give it time to come back again
      best = current;
    }
    
    // Kill bad downloads
    if (current != best) {
      long timeout;
      if (current->hasStarted()) timeout = global_config->candtryout;
      else timeout = global_config->candtimeout;

      if (current->getTimer() > timeout) {
	delete current;
      }
      else ++it;
    }
    else ++it;
  }

  /* If we havn't had enough candidates for more than NEWCANDIDATEDELAY
   * seconds, and we have search hits in our queue, promote a new search
   * hit to a download candidate.
   */
  if (now - newcandidatetime > global_config->canddelay && rungdl < global_config->dlcand) {
    /* We need to loop, the transfer may have been killed behind our backs*/
    while (!gdlqueue.isEmpty()) {
      QString gdlkey = gdlqueue.first();
      gdlqueue.pop_front();
      DcTransfer *next = getChildTransfer(gdlkey);
      if (next) {
	next->startTransfer();	
	newcandidatetime = now;
	break;
      }
    }
  }

  if (!best) {
    transferStatusChanged("Not found...");
    progress = target = -1;
    pos = size = speed = eta = -1; 
  }
  else {
    /* Put this data in us so updateStatusFromChildren get it */
    pos = best->getPos();
    size = best->getSize();
    speed = best->getSpeed();
    eta = best->getEta();
    progress = best->getProgress();
    target = best->getTarget();
    
    /* Show the user our status */
    transferStatusChanged("Downloading...",pos,size,speed,eta,progress,target);
  }

  /* See if we should do a new search as algorithm above */
  time_t timeout;
  if (gdlqueue.isEmpty()) timeout = global_config->nosourcetimeout;
  else if (stalled) timeout = global_config->stalledtimeout;
  else timeout = global_config->searchtimeout;

  // Is it time for a new search?
  if (now - lastsearch > timeout) performSearch();
}

void DcTrackTransfer::performSearch()
{
  QString modartist(getArtist());
  QString modalbum(getAlbum());
  QString modtitle(title);

  modartist = defunchyfy(modartist).replace(QRegExp("\\s"), "$");
  modalbum  = defunchyfy(modalbum).replace(QRegExp("\\s"), "$");
  modtitle  = defunchyfy(modtitle).replace(QRegExp("\\s"), "$");

  if (global_config->reqartist && global_config->reqalbum)
    getBackend()->performSearch(modartist+"$"+modalbum+"$"+modtitle);
  else if (global_config->reqartist)
    getBackend()->performSearch(modartist+"$"+modtitle);
  else if (global_config->reqalbum)
    getBackend()->performSearch(modalbum+"$"+modtitle);
  else
    getBackend()->performSearch(modtitle);

  transferStatusChanged("Queueing search",pos,size,speed,eta,progress,target);
  lastsearch = time(0);
}

bool DcTrackTransfer::modifyTransfer(QString & newkey)
{
  QRegExp trackrx("\\s*(\\d*)[^\\w]*([^\\[\\]]*)[^\\d]*(\\d*)[^\\d]*(\\d*)");
  trackrx.search(newkey);
  QString tmp1;
  QString tmp2;

  if (!(tmp1 = trackrx.cap(1)).isEmpty()) trackno = tmp1.toInt();
  if (!(tmp1 = trackrx.cap(2).stripWhiteSpace()).isEmpty()) title = tmp1;
  tmp1 = trackrx.cap(3);
  tmp2 = trackrx.cap(4);
  if (!tmp1.isEmpty() || !tmp2.isEmpty())
    length = (tmp1.isEmpty()?0:tmp1.toInt())*60 + (tmp2.isEmpty()?0:tmp2.toInt());

  newkey = formatKey(title, trackno, length);
  return DcTransfer::modifyTransfer(newkey);
}

/****************************/
/* DcProcess implementation */
/****************************/

DcProcess::DcProcess(DcBackend *parent, int key, DcHubConfig &conf, HubStatusView *hview) :
  QProcess(QString(DCTCBINARY)),
  parent(parent), key(key), dctclog(0), hview(hview), adress(conf.hubaddr),
  name(""), status("Connecting..."), users(0), downloads(0), uploads(0), totshares(0),
  uiticking(false), online(false)
{
  connect(this, SIGNAL(processExited()), this, SLOT(processDied()));
  connect(this, SIGNAL(readyReadStdout()), this, SLOT(slotReadyRead()));
  connect(this, SIGNAL(readyReadStderr()), this, SLOT(slotReadyReadStderr()));
  hview->addHub(key);
  connectHub(conf);
  updateStatus();
}

DcProcess::~DcProcess()
{
  /* Can only be killed from DcBackend, unregistered there... */
  disconnectHub();
  hview->killHub(key);
}

void DcProcess::connectHub(DcHubConfig &conf)
{
  QString tmp;
  if (isRunning()) disconnectHub();
  if (hview) hview->logText(key, "Starting client...");
  clearArguments();
  addArgument(DCTCBINARY);
  addArgument(QString("-g")+conf.hubaddr);
  
  addArgument(QString("-n")+conf.nick);
  addArgument(QString("-i")+conf.desc);
  addArgument(QString("-e")+conf.email);
  addArgument(QString("-c")+conf.conn);
  
  QStringList::iterator it;
  for (it = conf.shareddirs.begin(); it != conf.shareddirs.end(); ++it)
    addArgument(QString("-s")+*it);
 
  if (conf.usesizeoffs) {
    addArgument(QString("-o")+QString::number(conf.sizeoffs,'f',0));
  }
  
  if (conf.usevshare) addArgument(QString("-D")+conf.vsharedir);
  tmp.setNum(conf.maxupl);
  addArgument(QString("-d")+tmp);

  if (conf.useullim) {
    tmp.setNum(conf.ullim*2);
    addArgument(QString("-u")+tmp);
  }
  addArgument(QString("-b/LPATH ")+global_config->tempdir);

  if (global_config->activemode) {
    if (global_config->useoverrideip)
      addArgument(QString("-a")+global_config->overrideip);
    else if (global_config->dynamicip)
      addArgument("-b/DFLAG dynamic_ip 1");

    if (global_config->useoverrideport)
      addArgument(QString("-p")+QString::number(global_config->overrideport));
  }
  else addArgument("-f");

  if (global_config->useproxy) {
    addArgument(QString("-S")+global_config->proxyaddr);
    addArgument(QString("-P")+QString::number(global_config->proxyport));
    if (global_config->proxyversion == SOCKS5)
      addArgument("-x");

    if (!global_config->proxyuser.isEmpty()) {
      addArgument(QString("-U")+global_config->proxyuser);
      if (!global_config->proxypass.isEmpty()) {
	addArgument(QString("--sockspass=")+global_config->proxypass);
      }
    }
  }

  if (!start()) {
    processDied();
  }
  tmp.setNum(key);
  // dctclog = fopen("dctc-"+tmp+"-"+conf.name,"w");
}

void DcProcess::disconnectHub()
{
  sendRawCommand("/QUIT", "Closing hub connection");
  kill();
  if (dctclog) {
    fclose(dctclog);
    dctclog = 0;
  }
  status = "Offline";
  users = downloads = uploads = 0;
  totshares = 0;
  shares.clear();
  updateStatus();
}

void DcProcess::reconfigure(DcHubConfig &newconf)
{
  adress = newconf.hubaddr;
  connectHub(newconf);
}

void DcProcess::sendRawCommand(QString str, QString logtext)
{
  time_t now = time(0);
  char *cts = ctime(&now);
  char mytime[40];
  strcpy(mytime, cts);
  mytime[strlen(mytime)-1] = 0;
  if (online && isRunning()) {
    if (dctclog) {
      fprintf(dctclog,"[%s]<< %s\n",mytime,str.latin1());
      fflush(dctclog);
    }
    if (hview && !logtext.isEmpty()) hview->logText(key, logtext);
    writeToStdin(str+"\n");
  }
  else {
    if (dctclog) {
      fprintf(dctclog,"[%s]!< %s\n",mytime,str.latin1());
      fflush(dctclog);
    }
  }
}

void DcProcess::processDied()
{
  online = false;
  status = "Client exited";
  users = downloads = uploads = 0;
  totshares = 0;
  updateStatus();
  if (hview) hview->logText(key, "Client died, scheduling retry...");
  QTimer::singleShot(HUBRESTARTDELAY*1000,this,SLOT(restart()));
}

void DcProcess::restart()
{
  start();
}

void DcProcess::slotReadyRead()
{
  while (canReadLineStdout()) {
    QString tmp = partofline+readLineStdout();
    if (dctclog) {
      time_t now = time(0);
      char *cts = ctime(&now);
      char mytime[40];
      strcpy(mytime, cts);
      mytime[strlen(mytime)-1] = 0;
      fprintf(dctclog,"[%s]!> %s\n",mytime,tmp.latin1());
      fflush(dctclog);
    }
    parseOutput(tmp);
    partofline = "";
  }
  partofline += QString(readStdout());
}

void DcProcess::slotReadyReadStderr()
{
  while (canReadLineStderr()) {
    QString tmp = partoflinestderr+readLineStderr();
    
    if (dctclog) {
      time_t now = time(0);
      char *cts = ctime(&now);
      char mytime[40];
      strcpy(mytime, cts);
      mytime[strlen(mytime)-1] = 0;
      fprintf(dctclog,"[%s]!> %s\n",mytime,tmp.latin1());
      fflush(dctclog);
    }
    if (hview) hview->logText(key, tmp);
    partoflinestderr = "";
  }
  partoflinestderr += QString(readStderr());
}


void DcProcess::addGDL(long id, DcGDLTransfer *handler)
{
  handlers.insert(id, handler);
}

void DcProcess::removeGDL(long id)
{
  handlers.remove(id);
}

void DcProcess::parseOutput(QString line)
{
  static int tmpdownloads;

  time_t now = time(0);
  char *cts = ctime(&now);
  char mytime[40];
  strcpy(mytime, cts);
  mytime[strlen(mytime)-1] = 0;
  if (dctclog) {
    fprintf(dctclog,"[%s]>> %s\n",mytime,line.latin1());
    fflush(dctclog);
  }
  int pos;
  QRegExp mtyperx("(\\S*)\\s*\\]\\s*\"([^\"]*)\"");

  pos = mtyperx.search(line);
  pos += mtyperx.matchedLength();
  QString mtype = mtyperx.cap(1);
  QString debug = mtyperx.cap(2);

  if (mtype == 0 || debug == 0) {
    return;
  }

  QRegExp insrx("([^\\|]*)\\|");
  QValueVector<QString> inslist;

  while ((pos = insrx.search(line, pos)) != -1) {
    pos += insrx.matchedLength();
    inslist.push_back(insrx.cap(1));
  }

  // OK, we have data in mtype, debug and inslist, go make something useful

  if (mtype == "SREST") {
    if (hview) hview->logText(key, "Search result: " + inslist[1]);
    DcSearchResult hit(key, inslist[0], inslist[1], inslist[2].toLong());
    parent->searchHit(hit);
  }
  else if (mtype == "GLSTB") {
    tmpdownloads = 0;
    QIntDictIterator<DcGDLTransfer> it(handlers);
    DcGDLTransfer *cur;
    while ((cur = it.current())) {
      cur->gdlListingStart();
      ++it;
    }
  }
  else if (mtype == "GLSTC") {
    tmpdownloads++;
    long gdlid = inslist[0].toLong();
    DcGDLTransfer *curhand = handlers.find(gdlid);
    if (curhand) {
      long pos = inslist[4].toLong() - inslist[3].toLong();
      long speed = inslist[6].toLong();

      for (int listpos = 10; inslist[listpos] != ""; listpos++) {
	QRegExp substatex("([^$]*)\\$([^$]*)\\$([^$]*)\\$([^$]*)");
	
	substatex.search(inslist[listpos]);
	QString sub_user = substatex.cap(1);
	QString sub_path = substatex.cap(2);
	QString sub_size = substatex.cap(3);
	QString sub_stat = substatex.cap(4);

	QString id = sub_user + "|" + sub_path;
	long pos = 0;
	long start = -1;
	long end = -1;
	actiontype action;

	if (sub_stat == "W") action = Waiting;
	else if (sub_stat == "T") action = Trying;
	else {
	  action = Running;
	  QRegExp ranex("R(\\d*)\\[(\\d*);(\\d*)\\]");
	  ranex.search(sub_stat);
	  pos   = ranex.cap(1).toLong();
	  start = ranex.cap(2).toLong();
	  end   = ranex.cap(3).toLong();
	}
	curhand->setSourceStatus(id, action, pos, start, end);
      }
      curhand->setStatus(pos, speed);
    }
  }
  else if (mtype == "GLSTE") {
    if (tmpdownloads != downloads) {
      downloads = tmpdownloads;
      updateStatus();
    }
    QIntDictIterator<DcGDLTransfer> it(handlers);
    DcGDLTransfer *cur;
    while ((cur = it.current())) {
      cur->gdlListingEnd();
      /* Need to check if current have been deleted... 
	 if so, don't skip one item with ++it; */
      if (cur == it.current()) ++it;
    }
  }
  else if (mtype == "HUBNM") {
    name = inslist[0];
    if (hview) hview->logText(key, "Hubname set to: " + inslist[0]);
    updateStatus();
  }
  else if (mtype == "HUB+") {
    status = "Online";
    online = true;
    if (hview) hview->logText(key, "Hub is connected!");
    updateStatus();
  }
  else if (mtype == "HUB-") {
    status = "Offline";
    online = false;
    users = downloads = uploads = 0;
    totshares = 0;
    shares.clear();
    if (hview) hview->logText(key, "Hub has closed its connection!");
    updateStatus();
  }
  else if (mtype == "USER") {
    uiqueue.push_back(inslist[0]);
    if (!uiticking) {
      QTimer::singleShot(500,this,SLOT(sendUinfo()));
      uiticking = true;
    }
    if (hview) hview->logText(key, "New user entered the hub: " + inslist[0]);
  }
  else if (mtype == "USER+") {
    /* Don't do anything here, it will send UINFO */
    if (hview) hview->logText(key, "New user entered the hub: " + inslist[0]);
  }
  else if (mtype == "USER-") {
    if (hview) hview->logText(key, "User left the hub: " + inslist[0]);
    if (shares.contains(inslist[0])) {
      totshares -= shares[inslist[0]];
      shares.erase(inslist[0]);
      users--;
      updateStatus();
    }
    else uiqueue.remove(inslist[0]);
  }
  else if (mtype == "UINFO") {
    QStringList insparts = QStringList::split(' ',inslist[0],true);
    if (shares.contains(insparts[0]))
      totshares -= shares[insparts[0]];
    else
      users++;

    double ushares = insparts[1].toDouble();

    shares.insert(insparts[0],ushares);
    totshares += ushares;
    updateStatus();
  }
  else if (mtype == "$UL+") {
    uploads++;
    updateStatus();
  }
  else if (mtype == "$UL-") {
    uploads--;
    updateStatus();
  }
  else if (mtype == "$UL#") {
    uploads--;
    updateStatus();
  }
  else if (mtype == "CHAT") {
    QString tmp = inslist[0];
    tmp.replace("\r\r","\n");
    tmp.replace('\r',"\n");
    if (hview) hview->logText(key, tmp);
  }
  else if (mtype == "PRIV") {
    QString tmp = inslist[1];
    tmp.replace("\r\r","\n");
    tmp.replace('\r',"\n");
    if (hview) hview->logText(key, tmp);
  }
}

void DcProcess::sendUinfo()
{
  if (!uiqueue.isEmpty()) {
    sendRawCommand("/UINFO "+uiqueue.first());
    uiqueue.pop_front();
  }

  if (uiqueue.isEmpty()) {
    uiticking = false;
  }
  else {
    QTimer::singleShot(500,this,SLOT(sendUinfo()));
    uiticking = true;
  }
}
    

/****************************/
/* DcBackend implementation */
/****************************/

DcBackend::DcBackend(TransferAgent *rootagent, HubStatusView *hview) : 
  DcTransfer(rootagent, "Audio"),
  nextid(0), lastsearch(0), hview(hview)
{
  connect(&ticktimer, SIGNAL(timeout()), this, SLOT(genTick()));
  genTick();
}

DcBackend::~DcBackend() { deleteChildren(); }

void DcBackend::searchHit(DcSearchResult &result)
{
  QString match = result.path;

  /* The match has to end with some useful extension */
  if (match.find(QRegExp("(\\.mp3$)|(\\.ogg$)",false)) == -1) return;
  
  result.match = defunchyfy(match);
  DcTransfer::searchHit(result);
}

bool DcBackend::addHub(int key, DcHubConfig &conf)
{
  if (hubs.find(key)) return false;
  DcProcess *newproc = new DcProcess(this, key, conf, hview);
  hubs.insert(key, newproc);
  return true;
}

bool DcBackend::killHub(int key)
{
  DcProcess *p = hubs.take(key);
  if (!p) return false;
  delete p;
  return true;
}

bool DcBackend::reconfigureHub(int key, DcHubConfig &newconf)
{
  DcProcess *tmp;
  if (!(tmp = hubs.find(key))) return false;
  tmp->reconfigure(newconf);
  return true;
}

Transfer* DcBackend::generateAudioTransfer(QString artist, QString album, 
					   QString title,
					   int trackno, int length)
{
  DcArtistTransfer *arsl = dynamic_cast<DcArtistTransfer*>(getChildTransfer(artist));
  if (!arsl) arsl = new DcArtistTransfer(this, artist);

  DcAlbumTransfer  *alsl = dynamic_cast<DcAlbumTransfer*>(arsl->getChildTransfer(album));
  if (!alsl) alsl = new DcAlbumTransfer(arsl, album);

  if (alsl->getChildTransfer(title)) return 0;
  return new DcTrackTransfer(alsl, title, trackno, length);
}

void DcBackend::tick(time_t now)
{
  
  /* Things to do at regular intervals:
   * Check if there should be a new search.
   * List GDL:s
   * Pass the tick on
   */
  sendRawCommand("/GDLLST");
  if (now - lastsearch > DCSEARCHDELAY && !searchqueue.isEmpty()) {
    QString searchline = "/SRCH |"+searchqueue.first()+"|1";
    sendRawCommand(searchline,"Searching for \""+searchqueue.first()+"\"");
    searchqueue.pop_front();
    lastsearch = now;
  }
  DcTransfer::tick(now);
}

void DcBackend::genTick()
{
  tick(time(0));
  ticktimer.start(global_config->delay*1000,true);
}

/* Hardcore communication with dctc below */

void DcBackend::performSearch(QString findthis)
{
  if (!searchqueue.contains(findthis)) {
    searchqueue.append(findthis);
  }
}

long DcBackend::createGDL(int hub, long size, DcGDLTransfer *handler)
{
  DcProcess *proc = hubs.find(hub);

  proc->addGDL(nextid, handler);
  QString nextidstr = QString::number(nextid);
  QString sizestr   = QString::number(size);
  proc->sendRawCommand(QString("/GDLNEW ")+nextidstr+"|"+
		       sizestr+"|"+global_config->tempprefix+nextidstr,
		       "Creating new GDL with ID "+nextidstr);
  return nextid++;
}

void DcBackend::killGDL(int hub, long id)
{
  DcProcess *proc = hubs.find(hub);

  proc->removeGDL(id);
  QString idstr = QString::number(id);
  proc->sendRawCommand(QString("/GDLEND ")+idstr,
		       "Killing GDL with id "+idstr);
}

void DcBackend::createGDLSource(int hub, long id, 
				QString user, QString path, long size)
{
  DcProcess *proc = hubs.find(hub);
 
  QString idstr;
  idstr.setNum(id);
  QString sizestr;
  sizestr.setNum(size);
  proc->sendRawCommand("/GDLADD "+idstr+"|"+user+"|"+path+"|"+sizestr,
		       "Adding new GDL source with path "+path);
}

void DcBackend::killGDLSource(int hub, long id, QString user, QString path)
{
  DcProcess *proc = hubs.find(hub);

  QString idstr;
  idstr.setNum(id);
  proc->sendRawCommand(QString("/GDLDEL ")+idstr+"|"+user+"|"+path,
		       "Removing GDL source with path "+path);
}

void DcBackend::sendRawCommand(QString text, QString logtext)
{
  QIntDictIterator<DcProcess> it(hubs);
  DcProcess *current;
  while ((current = it.current())) {
    current->sendRawCommand(text, logtext);
    ++it;
  }
}
