#ifndef DCBACKEND_H
#define DCBACKEND_H

#include <ctime>
#include <iostream>

#include "qdict.h"
#include "qintdict.h"
#include "qprocess.h"
#include "qtimer.h"

#include "config.h"
#include "transfer.h"
#include "transferagent.h"
#include "transfergenerator.h"
#include "hubstatusview.h"
#include "config.h"

#define DCTCBINARY "dctc"
#define HUBRESTARTDELAY 20
#define DCSEARCHDELAY 5
#define GDLACTIVATETIMEOUT 10

using namespace std;
enum actiontype { Waiting, Trying, Running };

/***********************************************/
/* Classes used to filter incoming search hits */
/***********************************************/

class DcSearchResult
{
public:
  DcSearchResult(int hub, QString user, QString path, long size) :
    hub(hub), user(user), path(path), match(0), size(size) {}

  int hub;
  QString user;
  QString path;
  QString match;
  long size;
};

/*****************************************/
/* Helper master class for all transfers */
/*****************************************/

class DcTransfer : public Transfer
{
public:
  DcTransfer(TransferAgent *agent, QString key);
  DcTransfer(DcTransfer *parent, QString key, bool createagent = true);
  virtual ~DcTransfer();

  inline DcTransfer* getParentTransfer() { return parent; }
  inline DcTransfer* getChildTransfer(QString key) { return children.find(key); }
  inline int childrenCount() { return children.count(); }
  virtual void childCompleted(DcTransfer *child);

  virtual void searchHit(DcSearchResult &result);
  virtual void tick(time_t now);

  void updateStatusFromChildren();

  inline long   getPos()      { return pos; }
  inline long   getSize()     { return size; }
  inline double getProgress() { return progress; }
  inline double getTarget()   { return target; }
  inline long   getSpeed()    { return speed; }
  inline long   getEta()      { return eta; }
  inline long   isComplete()  { return complete; } 
  
  bool renameChild(QString oldkey, QString newkey)
  {
    DcTransfer *child = children.find(oldkey);
    if (!child) {
      cerr << "Woops! Submit a bug report!" << endl;
      cerr << "Didn't have child, I am " << key << endl;
      cerr << "Renaming from " << oldkey << " to " << newkey << endl;
      cerr << "Iterator listing: " << endl;
      QDictIterator<DcTransfer> it(children);
      for (; it.current(); ++it) {
	cerr << "  " << it.current()->key << endl;
      }
      cerr << "Listing ended.." << endl;
      return false; /* Degenerated case */
    }
    if (children.find(newkey)) return false;
    
    children.remove(oldkey);
    children.insert(newkey,child);
    return true;
  }
  static QString defunchyfy(QString str);

  /* Methods from Transfer, with helper implementation */
  virtual bool startTransfer();
  virtual bool stopTransfer();
  virtual bool resetTransfer();
  virtual bool abortTransfer() { delete this; return true; }
  virtual bool modifyTransfer(QString & newkey) {
    bool mayrename = true;
    if (parent) mayrename = parent->renameChild(key, newkey);
    if (mayrename) {
      key = newkey;
      return true;
    }
    return false;
  }
protected:
  inline void addChildTransfer(QString key, DcTransfer *child) { children.insert(key, child); }
  inline void removeChildTransfer(QString key) { children.remove(key); }
  void deleteChildren();

  QDict<DcTransfer> children;
  DcTransfer *parent;
  QString key;

  long   pos;
  long   size;
  double progress;
  double target;
  long   speed;
  long   eta;
  bool   complete;
  long   compsize;
  int    compcount;
};

/*******************************************/
/* DcProcess - deals with one dctc process */
/*******************************************/

class DcBackend;
class DcGDLTransfer;

class DcProcess : private QProcess
{
Q_OBJECT

public:
  DcProcess(DcBackend *parent, int key, DcHubConfig &conf, HubStatusView *hview);
  virtual ~DcProcess();
  void reconfigure(DcHubConfig &newconf);
  
  void addGDL(long id, DcGDLTransfer *handler);
  void removeGDL(long id);

  void sendRawCommand(QString str, QString logtext = "");

private slots:
  void processDied();
  void restart();
  void slotReadyRead();
  void slotReadyReadStderr();
  void sendUinfo();
private:
  inline void updateStatus() {
    if (hview) hview->updateHub(key, adress, name, status, users, downloads, uploads, totshares);
  }
  void parseOutput(QString line);
  void connectHub(DcHubConfig &conf);
  void disconnectHub();
  DcBackend *parent;
  int key;
  QIntDict<DcGDLTransfer> handlers;
  FILE *dctclog;
  HubStatusView *hview;
  QString adress;
  QString name;
  QString status;
  int users;
  int downloads;
  int uploads;
  double totshares;
  QMap<QString, double> shares;
  bool uiticking;
  QValueList<QString> uiqueue;
  bool online;
  QString partofline;
  QString partoflinestderr;
};


/***********************************************************/
/* Backend.. control over the connection and new transfers */
/***********************************************************/

class DcGDLTransfer;
class DcGDLSourceTransfer;

class DcBackend : public QObject, public TransferGenerator, public DcTransfer
{
Q_OBJECT

public:
  DcBackend(TransferAgent *rootagent, HubStatusView *hview);
  virtual ~DcBackend();
  /* Public controller methods */
  virtual bool addHub(int key, DcHubConfig &conf);
  virtual bool killHub(int key);
  virtual bool reconfigureHub(int key, DcHubConfig &newconf);
  /* TransferGenerator methods */
  virtual Transfer *generateAudioTransfer(QString artist, QString album, QString title, 
					  int trackno, int length);

  virtual void searchHit(DcSearchResult &result);
  /* Methods used internally to controll dctc */
  void performSearch(QString findthis);
  long createGDL(int hub, long size, DcGDLTransfer *handler);
  void killGDL(int hub, long id);
  void createGDLSource(int hub, long id, QString user, QString path, long size);
  void killGDLSource(int hub, long id, QString user, QString path);
private slots:
  void genTick();
private:
  virtual void tick(time_t now);

  void parseOutput(QString line);
  void sendRawCommand(QString text, QString logtext = "");

  QIntDict<DcProcess> hubs;
  QTimer ticktimer;
  FILE* dctclog;
  long nextid;
  time_t lastsearch;
  QValueList<QString> searchqueue;
  HubStatusView *hview;
};

class DcArtistTransfer : public DcTransfer
{
public:
  DcArtistTransfer(DcBackend *parent, QString artist);
  virtual ~DcArtistTransfer();
  virtual void searchHit(DcSearchResult &result);
  virtual void tick(time_t now);
  /* Info getting functions */
  DcBackend *getBackend() { return dynamic_cast<DcBackend*>(parent); }
  inline QString getArtist() { return key; }

  virtual bool allowesStartTransfer()  { return true; }
  virtual bool allowesStopTransfer()   { return true; }
  virtual bool allowesModifyTransfer() { return true; }
  virtual bool allowesResetTransfer()  { return true; }
  virtual bool allowesAbortTransfer()  { return true; }
};

class DcAlbumTransfer : public DcTransfer
{
public:
  DcAlbumTransfer(DcArtistTransfer *parent, QString album);
  virtual ~DcAlbumTransfer();
  virtual void searchHit(DcSearchResult &result);
  virtual void tick(time_t now);
  /* Info getting functions */
  inline DcBackend *getBackend() { return dynamic_cast<DcArtistTransfer*>(parent)->getBackend(); }
  inline QString getArtist() { return dynamic_cast<DcArtistTransfer*>(parent)->getArtist(); }
  inline QString getAlbum() { return key; }

  virtual bool allowesStartTransfer()  { return true; }
  virtual bool allowesStopTransfer()   { return true; }
  virtual bool allowesModifyTransfer() { return true; }
  virtual bool allowesResetTransfer()  { return true; }
  virtual bool allowesAbortTransfer()  { return true; }
};

/******************************************************/
/* DcTrackTransfer, associated with one song download */
/******************************************************/

class DcBackend;
class DcTrackTransfer : public DcTransfer
{
public:
  DcTrackTransfer(DcAlbumTransfer *parent,
		  QString title, int trackno, int length);
  virtual ~DcTrackTransfer();

  virtual void searchHit(DcSearchResult &result);
  virtual void tick(time_t now);
  virtual void performSearch();

  /* Info getting functions */
  inline DcBackend *getBackend() { return dynamic_cast<DcAlbumTransfer*>(parent)->getBackend(); }
  inline QString getArtist() { return dynamic_cast<DcAlbumTransfer*>(parent)->getArtist(); }
  inline QString getAlbum() { return dynamic_cast<DcAlbumTransfer*>(parent)->getAlbum(); }

  /* Functions that lets GDLTransfer update us */
  void gdlDone(DcGDLTransfer *gdl);
  inline void gdlStarted() { rungdl++; }
  inline void gdlStopped() { rungdl--; }
  inline void unregisterGDL(DcGDLTransfer *child) { if (child == best) best = 0; }
  /* Functions from Transfer */
  virtual bool startTransfer() { running = true;  DcTransfer::startTransfer(); return true; }
  virtual bool stopTransfer()  { running = false; DcTransfer::stopTransfer(); return true;  }
  virtual bool resetTransfer() { deleteChildren(); return true; }
  /*  virtual bool abortTransfer(); DcTransfer does it for us */
  virtual bool modifyTransfer(QString & newkey);
  /* Functions to show our capabilities */
  virtual bool allowesStartTransfer()  { return true; }
  virtual bool allowesStopTransfer()   { return true; }
  virtual bool allowesModifyTransfer() { return true; }
  virtual bool allowesResetTransfer()  { return true; }
  virtual bool allowesAbortTransfer()  { return true; }

private:
  static inline QString formatKey(QString title, int trackno, int length);
  QString title;
  int trackno;
  int length;
  bool running;
  QValueList<QString> gdlqueue;
  int rungdl;
  DcGDLTransfer *best;
  time_t lastsearch;
  time_t newcandidatetime;
};

/*********************************************/
/* Methods used to handle GDL and GDL source */
/*********************************************/

class DcGDLTransfer : public DcTransfer
{
public:
  DcGDLTransfer(DcTrackTransfer *parent, int hub, long size);
  virtual ~DcGDLTransfer();
  virtual void searchHit(DcSearchResult &result);
  /* Methods used internally */
  void gdlListingStart();
  void gdlListingEnd();
  void setStatus(long pos, long speed);
  void setSourceStatus(QString id, actiontype action, long pos, long start, long end);

  bool   hasStarted();
  void   resetTimer();
  time_t getTimer();

  /* Info getting functions */
  inline DcBackend *getBackend() { return dynamic_cast<DcTrackTransfer*>(parent)->getBackend(); }
  inline int getHub() { return hub; }
  inline long getGid() { return gid; }
  inline long getSize() { return size; }

  /* Functions to report status to TrackTransfer */
  inline QString getFilename() 
  { 
    QString gidstring;
    gidstring.setNum(gid);
    return global_config->tempdir+"/"+global_config->tempprefix+gidstring;
  }
  inline void gdlDone() { return dynamic_cast<DcTrackTransfer*>(parent)->gdlDone(this); }
  
  virtual bool startTransfer();
  virtual bool stopTransfer();
  virtual bool resetTransfer() { deleteChildren(); return true; }

  virtual bool allowesStartTransfer()  { return true; }
  virtual bool allowesStopTransfer()   { return true; }
  virtual bool allowesResetTransfer()  { return true; }
  virtual bool allowesAbortTransfer()  { return true; }

private:
  static QString intToString(long i);
  int hub;
  bool started;
  long gid;
  bool activated;
  bool hadgdl;
  bool nosources;
  bool stalled;
  time_t nosourcetime;
  time_t stalledtime;
  long lastpos;
  time_t timerstart;
  time_t starttime;
  time_t creationtime;
  QString filename;
};

class DcGDLSourceTransfer : public DcTransfer
{
public:
  DcGDLSourceTransfer(DcGDLTransfer *parent, 
		     QString user, QString path);
  virtual ~DcGDLSourceTransfer();
  void setStatus(actiontype action, long pos, long start, long end);

  /* Info getting functions */
  inline DcBackend *getBackend() { return dynamic_cast<DcGDLTransfer*>(parent)->getBackend(); }
  inline int getHub() { return dynamic_cast<DcGDLTransfer*>(parent)->getHub(); }
  inline long getGid()  { return dynamic_cast<DcGDLTransfer*>(parent)->getGid(); }
  inline long getSize() { return dynamic_cast<DcGDLTransfer*>(parent)->getSize(); }

  /* Transfer functions */
  virtual bool startTransfer();
  virtual bool stopTransfer();
  virtual bool resetTransfer();

  virtual bool allowesStartTransfer()  { return true; }
  virtual bool allowesStopTransfer()   { return true; }
  virtual bool allowesResetTransfer()  { return true; }
  virtual bool allowesAbortTransfer()  { return true; }

private:
  QString user;
  QString path;

  bool started;
  long lastpos;
  bool stalled;
  int attempts;
  time_t stalledtime;
};

#endif // DCBACKEND_H

