#include "qlistview.h"
#include "qpopupmenu.h"

#include "downloadview.h"
#include "transfer.h"
#include "transferagent.h"


/*******************************/
/* DownloadItem implementation */
/*******************************/

DownloadItem::DownloadItem(DownloadView *parent, bool newtransfer) : 
  PaintListViewItem(parent), TransferAgent(),
  progressbar(new PaintProgress(this, 1000)),
  completed(false), createOnRename(newtransfer)
{
  addPaintWidget(4,progressbar);
  if (newtransfer) {
    setRenameEnabled(0, true);
  }
  else {
    setText(0,"New...");
  }
}

DownloadItem::DownloadItem(DownloadItem *parent, bool newtransfer) : 
  PaintListViewItem(parent), TransferAgent(),
  progressbar(new PaintProgress(this,1000)),
  completed(false), createOnRename(newtransfer)
{
  if (newtransfer) {
    setRenameEnabled(0, true);
  }
  addPaintWidget(4,progressbar);
}

DownloadItem::~DownloadItem()
{
  /* We don't have to do any local cleanup */
}

void DownloadItem::startRename(int col)
{
  bool renameable = allowesModifyTransfer();
  setRenameEnabled(0,renameable);
  QListViewItem::startRename(col);
  if (renameable) dynamic_cast<DownloadView*>(listView())->getNoupdateLock();
}

void DownloadItem::transferCreated(Transfer *transfer, QString id)
{
  TransferAgent::transferCreated(transfer, id);
  setText(0, id);
  key = id;
}

void DownloadItem::transferStatusChanged(QString status, long pos, long size,
					 long speed, long eta,
					 double progress, double target) 
{
  setText(1, status);
  setText(2, formatSize(pos));
  setText(3, formatSize(size));
  setDualProgress(progress, target);
  setText(5, formatSpeed(speed));
  setText(6, formatTime(eta));
}

void DownloadItem::transferCompleted()
{
  setText(1, "Completed!");
  completed = true;
}

void DownloadItem::transferRemoved()
{
  TransferAgent::transferRemoved();
  if (!completed) setText(1, "Aborted!");
  delete this;
}

TransferAgent *DownloadItem::transferChildCreated()
{
  return new DownloadItem(this);
}

/* Methods used for formatting and displaying */

QString DownloadItem::formatSize(long size)
{
  QString str;

  if (size < 0) str = "";
  else if (size >= 1.0*1024*1024*1024*10)
    str.sprintf("%.1fGB", size / (1.0*1024*1024*1024));
  else if (size >= 1.0*1024*1024*1024)
    str.sprintf("%.2fGB", size / (1.0*1024*1024*1024));
  else if (size >= 1.0*1024*1024*10)
    str.sprintf("%.1fMB", size / (1.0*1024*1024));
  else if (size >= 1.0*1024*1024)
    str.sprintf("%.2fMB", size / (1.0*1024*1024));
  else if (size >= 1.0*1024*10)
    str.sprintf("%.1fKB", size / (1.0*1024));
  else if (size >= 1.0*1024)
    str.sprintf("%.2fKB", size / (1.0*1024));
  else if (size >= 1.0*10)
    str.sprintf("%.1fB", size * 1.0);
  else 
    str.sprintf("%.2fB", size * 1.0);

  return str;
}

QString DownloadItem::formatSpeed(long speed)
{
  QString str;

  if (speed < 0) str = "";
  else if (speed >= 1.0*1024*1024*1024*10)
    str.sprintf("%.1fGB/s", static_cast<double>(speed) / (1.0*1024*1024*1024));
  else if (speed >= 1.0*1024*1024*1024)
    str.sprintf("%.2fGB/s", static_cast<double>(speed) / (1.0*1024*1024*1024));
  else if (speed >= 1.0*1024*1024*10)
    str.sprintf("%.1fMB/s", static_cast<double>(speed) / (1.0*1024*1024));
  else if (speed >= 1.0*1024*1024)
    str.sprintf("%.2fMB/s", static_cast<double>(speed) / (1.0*1024*1024));
  else if (speed >= 1.0*1024*10)
    str.sprintf("%.1fKB/s", static_cast<double>(speed) / (1.0*1024));
  else if (speed >= 1.0*1024)
    str.sprintf("%.2fKB/s", static_cast<double>(speed) / (1.0*1024));
  else if (speed >= 1.0*10)
    str.sprintf("%.1fB/s", static_cast<double>(speed));
  else if (speed > 0)
    str.sprintf("%.2fB/s", static_cast<double>(speed));
  else 
    str = "-stalled-";

  return str;
}

QString DownloadItem::formatPercentage(double frac)
{
  QString str;
  if (frac < 0) str = "";
  else str.sprintf("%i%%", (int)(frac*100));
  return str;
}

QString DownloadItem::formatTime(long sec)
{
  // Special values used to mark no ETA or infinite time
  if (sec < 0) return "";

  QString str;
  long h = sec / 60 / 60;
  int  m = sec / 60 % 60;
  int  s = sec % 60 % 60;

  if (h)
    str.sprintf("%lih%im%is", h, m, s);
  else if (m)
    str.sprintf("%im%is", m, s);
  else
    str.sprintf("%is", s);

  return str;
}

void DownloadItem::setProgress(double prog)
{
  progressbar->setProgress(static_cast<int>(prog*1000));
}

void DownloadItem::setDualProgress(double p1, double p2)
{
  progressbar->setProgress(static_cast<int>(p1*1000));
  progressbar->setProgress2(static_cast<int>(p2*1000));
}

void DownloadItem::doPopupMenu(const QPoint &pos)
{
  QPopupMenu pop(listView());

  if (allowesCreateTransfer()) pop.insertItem("New",  0);
  if (allowesStartTransfer())  pop.insertItem("Start",1);
  if (allowesStopTransfer())   pop.insertItem("Stop", 2);
  if (allowesResetTransfer())  pop.insertItem("Reset",3);
  if (allowesAbortTransfer())  pop.insertItem("Abort",4);

  switch (pop.exec(pos)) {
  case 0: new DownloadItem(this, true); break;
  case 1: startTransfer(); break;
  case 2: stopTransfer();  break;
  case 3: resetTransfer(); break;
  case 4: abortTransfer(); break;
  }
}

void DownloadItem::okRename(int col)
{
  QListViewItem::okRename(col);
  dynamic_cast<DownloadView*>(listView())->dropNoupdateLock();
  if (createOnRename) {
    if (createTransfer(this, text(0))) {
      createOnRename = false;
    }
    else {
      delete this;
      return;
    }
  }
  else {
    QString tmp = text(0);
    if (modifyTransfer(tmp)) key = tmp;
    setText(0,key);
  }
}

void DownloadItem::cancelRename(int col)
{
  QListViewItem::cancelRename(col);
  dynamic_cast<DownloadView*>(listView())->dropNoupdateLock();
}
  
/*******************************/
/* DownloadView implementation */
/*******************************/

DownloadView::DownloadView(QWidget *parent) : QListView(parent), noupdatelock(0)
{
  addColumn("Name");
  addColumn("Status");
  addColumn("Position");
  addColumn("Size");
  addColumn("Progress");
  addColumn("Speed");
  addColumn("ETA");
  setColumnWidthMode(0,Manual);
  setColumnWidthMode(1,Manual);
  setColumnWidthMode(2,Manual);
  setColumnWidthMode(3,Manual);
  setColumnWidthMode(4,Manual);
  setColumnWidthMode(5,Manual);
  setColumnWidthMode(6,Manual);
  setColumnWidth(0,200);
  setColumnWidth(1,80);
  setColumnWidth(2,50);
  setColumnWidth(3,50);
  setColumnWidth(4,80);
  setColumnWidth(5,50);
  setColumnWidth(6,50);

  setRootIsDecorated(true);

  connect(this, SIGNAL(contextMenuRequested(QListViewItem*, const QPoint &, int)),
	  this, SLOT(doPopupMenu(QListViewItem*, const QPoint &, int)));
}

DownloadView::~DownloadView()
{
}

void DownloadView::transferCreated(Transfer *transfer, QString id) 
{
  /* This should never happen at this level */
  TransferAgent::transferCreated(transfer, id);
}

void DownloadView::transferStatusChanged(QString status , long pos, long size,
					 long speed, long eta,
					 double progress, double target)
{
  static_cast<void>(status); /* Warning eater */
  static_cast<void>(pos); /* Warning eater */
  static_cast<void>(size); /* Warning eater */
  static_cast<void>(speed); /* Warning eater */
  static_cast<void>(eta); /* Warning eater */
  static_cast<void>(progress); /* Warning eater */
  static_cast<void>(target); /* Warning eater */

  /* Possibly show info of total download speed or something here */
}

void DownloadView::transferCompleted()
{
  /* This should never happen at this level */
}

void DownloadView::transferRemoved()
{
  /* This should never happen at this level */
}

TransferAgent *DownloadView::transferChildCreated()
{
  return new DownloadItem(this);
}

void DownloadView::doPopupMenu(QListViewItem *item, const QPoint &pos, int col)
{
  static_cast<void>(col); /* Warning eater */
  if (item) dynamic_cast<DownloadItem*>(item)->doPopupMenu(pos);
}
